% askisi_09.pl (Encoding: UTF-8)
% Εργαστηριακές Ασκήσεις Τεχνητής Νοημοσύνης με τη Γλώσσα Prolog
% Κυριάκος Σγάρμπας
% 2024, Κάλλιπος, Ανοικτές Ακαδημαϊκές Εκδόσεις
% ------------------------------------------------------------------------------
% ΑΣΚΗΣΗ 9: ΑΛΓΟΡΙΘΜΟΙ ΑΝΑΖΗΤΗΣΗΣ Ι
% ------------------------------------------------------------------------------
% Ερώτημα 1
% ------------------------------------------------------------------------------

% Breadth-First Search
% Καλείται ως: find_BFS(+State1,+State2,-Node)
search_BFS([G|_],G):- !.
search_BFS([CurrentNode|Fringe],GoalNode):-
    findall(SuccNode,successor(CurrentNode,SuccNode),NList),
    append(Fringe,NList,NewFringe), search_BFS(NewFringe,GoalNode).
find_BFS(Initial_State,Goal_State,GoalNode):-
    GoalNode=node(_,Goal_State,_,_,_),
    search_BFS([node(0,Initial_State,move,0,null)],GoalNode).

% Συνάρτηση Διαδόχων
successor(NODE,SNODE):-
    NODE=node(PathCost, State, Action, Depth, _),
    SNODE=node(SPathCost,SState,SAction,SDepth,SParent),
    next(State,SState,Cost), SAction=Action,  % Εξαρτώνται από το πρόβλημα.
    SDepth is Depth+1, SPathCost is PathCost+Cost, SParent=NODE.

% Δεδομένα από το πρόβλημα του οδικού χάρτη
link(a,z,75).  link(a,s,140). link(a,t,118). link(z,o,71).  link(o,s,151).
link(t,l,111). link(l,m,70).  link(m,d,75).  link(d,c,120). link(c,r,146).
link(c,p,138). link(s,r,80).  link(r,p,97).  link(p,b,101). link(s,f,99).
link(f,b,211). link(b,g,90).  link(u,h,98).  link(h,e,86).  link(u,v,142).
link(v,i,92).  link(n,i,87).

next(X,Y,D):- link(X,Y,D); link(Y,X,D).

/*
?- find_BFS(a,b,N).
N = node(450,b,move,3,node(239,f,move,2,node(140,s,move,1,node(0,a,move,0,null))))
*/

% ------------------------------------------------------------------------------
% Ερώτημα 2
% ------------------------------------------------------------------------------

% Uniform Cost Search
% Καλείται ως: find_UCS(+State1,+State2,-Node)
search_UCS([G|_],G):- !.
search_UCS([CurrentNode|Fringe],GoalNode):-
    findall(SuccNode,successor(CurrentNode,SuccNode),NList),
    append(Fringe,NList,NewFringe), msort(NewFringe,SNFringe),
    search_UCS(SNFringe,GoalNode).
find_UCS(Initial_State,Goal_State,GoalNode):-
    GoalNode=node(_,Goal_State,_,_,_),
    search_UCS([node(0,Initial_State,move,0,null)],GoalNode).

/*
?- find_UCS(a,b,N).
N = node(418,b,move,4,node(317,p,move,3,node(220,r,move,2,node(140,s,move,1,node(0,a,move,0,null)))))
*/

% ------------------------------------------------------------------------------
% Ερώτημα 3
% ------------------------------------------------------------------------------

% Depth-First Search
% Κλήση: find_DFS(+State1,+State2,-Node)
search_DFS(G,G):- !.
search_DFS(CurrentNode,GoalNode):-
    successor(CurrentNode,SuccNode), search_DFS(SuccNode,GoalNode).
find_DFS(Initial_State,Goal_State,GoalNode):-
    GoalNode=node(_,Goal_State,_,_,_),
    search_DFS(node(0,Initial_State,move,0,null),GoalNode), !.

/*
?- find_DFS(a,b,N).
N = node(575,b,move,6,node(474,p,move,5,node(377,r,move,4,node(297,s,move,3,
    node(146,o,move,2,node(75,z,move,1,node(0,a,move,0,null)))))))

?- find_DFS(i,n,N).
Stack limit (0.2Gb) exceeded
  Stack sizes: local: 0.1Gb, global: 35.2Mb, trail: 15.1Mb
  Stack depth: 328,941, last-call: 0%, Choice points: 328,918
  Possible non-terminating recursion:
    [328,938] search_DFS(<compound node/5>, <compound node/5>)
    [328,937] search_DFS(<compound node/5>, <compound node/5>)
*/

% Depth-Limited Search (Depth-First Search Περιορισμένου Βάθους)
% Κλήση: find_DLS(+State1,+State2,+Limit,-Node)
search_DLS(G,G,_):- !.
search_DLS(CurrentNode,GoalNode,Limit):-
    CurrentNode=node(_,_,_,D,_), D<Limit,
    successor(CurrentNode,SuccNode),
    search_DLS(SuccNode,GoalNode,Limit).
find_DLS(Initial_State,Goal_State,Limit,GoalNode):-
    GoalNode=node(_,Goal_State,_,_,_),
    search_DLS(node(0,Initial_State,move,0,null),GoalNode,Limit), !.

/*
?- find_DLS(a,b,7,N).
N = node(575,b,move,6,node(474,p,move,5,node(377,r,move,4,node(297,s,move,3,
    node(146,o,move,2,node(75,z,move,1,node(0,a,move,0,null)))))))

?- find_DLS(a,b,5,N).
N = node(607,b,move,5,node(396,f,move,4,node(297,s,move,3,node(146,o,move,2,
    node(75,z,move,1,node(0,a,move,0,null))))))

?- find_DLS(a,b,4,N).
N = node(418,b,move,4,node(317,p,move,3,node(220,r,move,2,node(140,s,move,1,
    node(0,a,move,0,null)))))

?- find_DLS(a,b,3,N).
N = node(450,b,move,3,node(239,f,move,2,node(140,s,move,1,node(0,a,move,0,null))))

?- find_DLS(a,b,2,N).
false

?- find_DLS(i,n,15,N).
N = node(1375,n,move,15,node(1288,i,move,14,node(1196,v,move,13,node(1104,i,
    move,12,node(1012,v,move,11,node(920,i,move,10,node(828,v,move,9,
    node(736,i,move,8,node(644,v,move,7,node(552,i,move,6,
    node(460,v,move,5,node(368,i,move,4,node(276,v,move,3,
    node(184,i,move,2,node(92,v,move,1,node(0,i,move,0,null))))))))))))))))

?- find_DLS(i,n,5,N).
N = node(455,n,move,5,node(368,i,move,4,node(276,v,move,3,node(184,i,move,
    2,node(92,v,move,1,node(0,i,move,0,null))))))

?- find_DLS(i,n,4,N).
N = node(271,n,move,3,node(184,i,move,2,node(92,v,move,1,node(0,i,move,0,null))))

?- find_DLS(i,n,2,N).
N = node(87,n,move,1,node(0,i,move,0,null))
*/

% ------------------------------------------------------------------------------
% Ερώτημα 4
% ------------------------------------------------------------------------------

% Iterative Deepening Search
% Κλήση: find_IDS(+State1,+State2,-Node)
search_IDS(CurrentNode,GoalNode,D):-
    write('Depth = '), write(D), nl,            % Για οπτικό έλεγχο λειτουργίας.
    search_DLS(CurrentNode,GoalNode,D), !.
search_IDS(CurrentNode,GoalNode,D):-
    NewD is D+1, search_IDS(CurrentNode,GoalNode,NewD).
find_IDS(Initial_State,Goal_State,GoalNode):-
    GoalNode=node(_,Goal_State,_,_,_),
    search_IDS(node(0,Initial_State,move,0,null),GoalNode,0), !.

/*
?- find_IDS(a,b,N).
Depth = 0
Depth = 1
Depth = 2
Depth = 3
N = node(450,b,move,3,node(239,f,move,2,node(140,s,move,1,node(0,a,move,0,null))))

?- find_IDS(n,i,N).
Depth = 0
Depth = 1
N = node(87,i,move,1,node(0,n,move,0,null))
*/
